#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <gtk/gtk.h>
#include <gdk/gdkx.h>
#include <stdio.h>
#include <X11/extensions/XTest.h>
#include <X11/keysym.h>
#include <ctype.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <locale.h>


#include "../libdictionary/predictive_dictionary.h"


#define MAX_WORD 30
struct word_info {

    char* word;
    int end;

} word[ MAX_WORD ];
int words;

GtkWidget* window;
GtkWidget* prediction_label;
GtkWidget* cycle_button;
GtkWidget* event_box;

char input_string[200];
char* predicted_word = NULL;

PredictiveDictionary* dict;
Display *dpy;    
Window win;
Window last_win = -1;
volatile int sending_keys;


unsigned long x_mask = KeyPressMask|SubstructureNotifyMask|FocusChangeMask;
void select_all_windows(Display *disp, Window root, unsigned long type); 


void update_word( char* word ) {

    char buffer[200];
    int len;
    char* label;

    if( word == NULL )
        gtk_label_set_text( GTK_LABEL( prediction_label ), "" );
    else {

        len =strlen( input_string );

        if( len > 0 ) {

            sprintf( buffer, "<u>%s</u>%s", input_string, &word[ len ] ); 
            label = g_locale_to_utf8( buffer, -1, NULL, NULL, NULL ); 
            gtk_label_set_markup( GTK_LABEL( prediction_label ), label );
            g_free( label );

        }
        else {

            label = g_locale_to_utf8( word, -1, NULL, NULL, NULL ); 
            gtk_label_set_text( GTK_LABEL( prediction_label ), label );
            g_free( label );

        }

    }

}

void complete_word( char* predicted_word ) {

    char* insert_word;
    KeySym ks;
    char key_string[2] = { 0, '\0' };
    int size, i;

    size = strlen( input_string ); 


    if( predicted_word != NULL && size < strlen( predicted_word ) ) {

        if( last_win != -1 && last_win != win ) {

            XSetInputFocus( dpy, last_win, RevertToNone, CurrentTime );
            insert_word = &predicted_word[ size ];

            // send word
            sending_keys = 1;
            for( i = 0; insert_word[i]; i++ ) {

                key_string[0] = insert_word[i];
                XTestFakeKeyEvent(dpy, XKeysymToKeycode( dpy, XStringToKeysym( key_string ) ), TRUE, 0 );
                XTestFakeKeyEvent(dpy, XKeysymToKeycode( dpy, XStringToKeysym( key_string ) ), FALSE, 0 );

            }
            XSync( dpy, True );
            sending_keys = 0;
            XSync( dpy, False );

        }
    }

}



void clear_prediction() {

    gtk_label_set_text( GTK_LABEL( prediction_label ), "" );
    predicted_word == NULL;
    input_string[0] = '\0';
    words = 0;

}

void update_prediction( int cycle ) {

    char* new_word;
    int width, dummy;
    char display_string[150];
    PangoLayout* layout;
    int size, length, word_size, i;
    char* utf8_crap;

    if( cycle )
        predicted_word = predictive_dictionary_cycle( dict, input_string );
    else
        predicted_word = predictive_dictionary_predict( dict, input_string );

    if( predicted_word != NULL ) {

        gdk_drawable_get_size( event_box->window, &width, &dummy );

        layout = pango_layout_copy( gtk_label_get_layout( GTK_LABEL( prediction_label ) ) );

        strcpy( display_string, predicted_word );
        length = strlen( predicted_word );
        utf8_crap = g_locale_to_utf8( display_string, -1, NULL, NULL, NULL ); 
        pango_layout_set_text( layout, utf8_crap, -1 );
        g_free( utf8_crap );
        pango_layout_get_pixel_size( layout, &word_size, &dummy);
        size = word_size; 

        word[0].word =predicted_word;
        word[0].end = word_size;
        i = 1;
        while( i < MAX_WORD ) {

            new_word = predictive_dictionary_cycle( dict, input_string );

            if( new_word == NULL || new_word == predicted_word )
                break;

            length += strlen( new_word );
            utf8_crap = g_locale_to_utf8( new_word, -1, NULL, NULL, NULL ); 
            pango_layout_set_text( layout, utf8_crap, -1 );
            g_free( utf8_crap );
            pango_layout_get_pixel_size( layout, &word_size, &dummy);
            size += word_size + 3; // compensate for the space 

            if( length > 150 || size > width )
                break;

            word[i].word = new_word;
            word[i].end = size;

            strcat( display_string, " " );
            strcat( display_string, new_word );

            i++;

        }
        words = i;


        update_word( display_string );


    }
    else if( input_string[0] != '\0' )
        update_word( input_string );
    else
        clear_prediction();

}

void cycle_button_clicked( GtkWidget* button, gpointer data ) {

    update_prediction( 1 );

}

int word_is_valid( char* word ) {

    char* i = word;
    while( *i && isalpha( *i ) )
        i++;

    return ( *i ? 0 : 1 );

}


GdkFilterReturn key_press_filter(GdkXEvent *xevent, GdkEvent *event, gpointer data) {

    static int count;
    static char buffer[200];
    static KeySym ks;
    int len;
    XCreateWindowEvent *xcwe;


    if ( ( (XEvent*) xevent)->type==CreateNotify) {

        XSync( dpy, True );
        xcwe=(XCreateWindowEvent *)xevent;
        select_all_windows(dpy, DefaultRootWindow(dpy) /* xcwe->window */, x_mask );
        XSync( dpy, False );

    }
    else if ( ( (XEvent*) xevent)->type== FocusOut) {

        clear_prediction();
        
    }
    else if( ( (XEvent*) xevent )->type == KeyPress && !sending_keys) {

        count = XLookupString((XKeyEvent *) xevent, buffer, 200, &ks,NULL);

        if( count > 0  ) { 

           if( buffer[0] == 8  ) {

                len = strlen( input_string );
                if( len > 1 ) {

                    input_string[ len - 1 ] = '\0';
                    update_prediction( 0 );

                }
                else
                    clear_prediction();

            }
            else if ( word_is_valid( buffer ) ) {

                last_win = ( (XKeyEvent*) xevent )->window;
                strcat( input_string, buffer );
                update_prediction( 0 );

            }
            else { //if( buffer[0] == 13 || buffer[0] == 32  ) {

                predictive_dictionary_update_with_word( dict, input_string );
                clear_prediction();


            }
 
        }

    }

    return GDK_FILTER_CONTINUE;

}

gboolean prediction_label_button_press( GtkWidget* widget, GdkEventButton* event, gpointer data ) {

    int i, width, height;


    for( i = 0; i < words && word[i].end < (int) event->x; i++ )
        ;

    if( i != words ) {
      
        if( word[i].word ) {

            complete_word( word[i].word );
            strcpy( input_string, word[i].word );
            update_prediction( 0 );

        }

    }

}

void clean_up() {

    predictive_dictionary_destroy( dict );

}

void signal_handler( int sig_num ) {

    if( sig_num == SIGTERM ) {

        clean_up();

    }

}


/*
 * A dummy error handler for X errors, as the may occur when accesssing a BadWindow
 *
 */
int error_handler( Display* disp, XErrorEvent* error ) {

    return 0;

}

gboolean window_delete( GtkWidget* widget, GdkEvent *event, gpointer data ) {

    clean_up();

    return FALSE;
}


void configure_dictionary() {

    char *home_dir, *common_dict_filename, *static_dict_filename = NULL;
    char filename[200];
    struct stat buf;

    home_dir = getenv( "HOME" );
    if( home_dir ) {
        
        snprintf( filename, 100, "%s/.minipredict_common_dictionary", home_dir );
        common_dict_filename = strdup( filename );

        snprintf( filename, 100, "%s/.minipredict_dictionary", home_dir );

        if( stat( filename, &buf ) == 0 ) {

            if( buf.st_size  > 0 )
                static_dict_filename = strdup( filename );

        }

    }
    else {

        strcpy( filename, "/tmp/common_dictionary" );
        common_dict_filename = strdup( filename );

    }

    if( static_dict_filename == NULL )
        static_dict_filename = strdup( "/usr/share/dictionary/dictionary" );

    dict = predictive_dictionary_new( common_dict_filename, static_dict_filename );
    input_string[0] = '\0';

    free( static_dict_filename );
    free( common_dict_filename );

}

int main( int argc, char** argv ) {

    int status;
    GtkWidget* hbox;
    Atom window_type_atom, window_type_toolbar_atom;
    XWMHints *wm_hints;

    
    // catch SIGTERM
    signal( SIGTERM, signal_handler );


    gtk_init( &argc, &argv );


    // set up the window. code copied from gpe-nmf
    window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
    gtk_signal_connect( GTK_OBJECT( window ), "delete-event",
                        GTK_SIGNAL_FUNC( window_delete ), NULL );
    gtk_window_set_title (GTK_WINDOW (window), "Word completion");
    gtk_widget_set_usize (GTK_WIDGET (window), gdk_screen_width(), 17);
    gtk_container_set_border_width( GTK_CONTAINER( window ), 0 );
    gtk_widget_realize(window);
    win = gdk_x11_drawable_get_xid( window->window );

    dpy = GDK_WINDOW_XDISPLAY (window->window);
    window_type_atom =
        XInternAtom (dpy, "_NET_WM_WINDOW_TYPE" , False);
    window_type_toolbar_atom =
        XInternAtom (dpy, "_NET_WM_WINDOW_TYPE_TOOLBAR",False);
    XChangeProperty (dpy, GDK_WINDOW_XWINDOW (window->window), 
            window_type_atom, XA_ATOM, 32, PropModeReplace, 
            (unsigned char *) &window_type_toolbar_atom, 1);

    hbox = gtk_hbox_new( FALSE, 0 );

    event_box = gtk_event_box_new();
    gtk_signal_connect( GTK_OBJECT( event_box ), "button-press-event",
            GTK_SIGNAL_FUNC( prediction_label_button_press ), NULL );
    prediction_label = gtk_label_new( "" );
    gtk_misc_set_alignment( GTK_MISC( prediction_label ), 0, 0 );
    gtk_container_add( GTK_CONTAINER( event_box ), prediction_label );
    gtk_container_set_border_width( GTK_CONTAINER( event_box ), 2 );
    gtk_box_pack_start( GTK_BOX( hbox ), event_box, TRUE, TRUE, 0 );

    cycle_button = gtk_button_new();
    gtk_button_set_relief( GTK_BUTTON( cycle_button ), GTK_RELIEF_NONE );
    gtk_container_set_border_width( GTK_CONTAINER( cycle_button ), 0 );
    gtk_container_add( GTK_CONTAINER( cycle_button), gtk_image_new_from_stock( GTK_STOCK_GO_FORWARD, GTK_ICON_SIZE_MENU ) );
    gtk_signal_connect( GTK_OBJECT( cycle_button ), "clicked",
            GTK_SIGNAL_FUNC( cycle_button_clicked ), NULL );
    gtk_box_pack_start( GTK_BOX( hbox ), cycle_button, FALSE, FALSE, 0 );

    gtk_container_add( GTK_CONTAINER( window ), hbox );
    gtk_widget_show_all( window );


    // set up the dictionary
    configure_dictionary();
    // set up X stuff
    wm_hints = XAllocWMHints();
    wm_hints->input = False;
    wm_hints->flags = InputHint;
    XSetWMHints(dpy, win, wm_hints );

    XSetErrorHandler( error_handler );

    XSync( dpy, True );
    select_all_windows(dpy, DefaultRootWindow(dpy), x_mask );
    XSync( dpy, False );

    // add a filter to catch keypress etc. 
    gdk_window_add_filter( NULL, key_press_filter, NULL );

    gtk_main();

    return 0;

}



void select_all_windows(Display *disp, Window root, unsigned long type) {

    Window	parent, *children;
    int	stat, i, j, k;
    unsigned int	nchildren;

    if( root == win )
        return;

    XSelectInput(disp, root, type);

    stat=XQueryTree(disp, root, &root, &parent, &children, &nchildren);
    if (stat == FALSE) {
//        fprintf(stderr, "Can't query window tree.\n");
        return;
    }
    if (nchildren == 0) return;

    /* select for all children windows too */
    for(i=0; i < nchildren; i++) {
        select_all_windows(disp, children[i], type);
    }     

    XFree((char *)children);

}

